<?php

declare(strict_types=1);

namespace Drupal\data_pipelines\Index;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\data_pipelines\DatasetData;
use Drupal\data_pipelines\Entity\DatasetInterface;
use Drupal\elasticsearch_connector\ClusterManager;
use Drupal\elasticsearch_connector\ElasticSearch\ClientManager;
use Drupal\elasticsearch_connector\Entity\Cluster;

/**
 * Defines a class for indexing data to elastic search.
 */
final class DatasetElasticsearchIndexer implements DatasetIndexerInterface {

  const INDEX_ID = 'data_pipelines';
  const INDEX_SHARDS = 5;
  const INDEX_REPLICAS = 1;
  const INDEX_CODEC = 'default';

  /**
   * Client.
   *
   * @var \nodespark\DESConnector\ClientInterface|\Elasticsearch\Client
   */
  protected $client;

  /**
   * Logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $loggerChannel;

  /**
   * The module config.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Factory method.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity type manager.
   * @param \Drupal\elasticsearch_connector\ElasticSearch\ClientManager $client_manager
   *   Client manager.
   * @param \Drupal\elasticsearch_connector\ClusterManager $cluster_manager
   *   Cluster manager.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger_channel
   *   Logger.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   *
   * @return \Drupal\data_pipelines\Index\DatasetElasticsearchIndexer
   *   New instance.
   */
  public static function factory(EntityTypeManagerInterface $entity_type_manager, ClientManager $client_manager, ClusterManager $cluster_manager, LoggerChannelInterface $logger_channel, ConfigFactoryInterface $config_factory): DatasetElasticsearchIndexer {
    $cluster = $entity_type_manager->getStorage('elasticsearch_cluster')->load($cluster_manager->getDefaultCluster());
    assert($cluster instanceof Cluster);
    $client = $client_manager->getClientForCluster($cluster);
    $config = $config_factory->get('data_pipelines.settings');
    return new static($client, $logger_channel, $config);
  }

  /**
   * DatasetStorage constructor.
   *
   * @param \nodespark\DESConnector\ClientInterface|\Elasticsearch\Client $client
   *   Elastic search client.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger_channel
   *   Logger.
   * @param \Drupal\Core\Config\ImmutableConfig $config
   *   The module configuration.
   */
  public function __construct($client, LoggerChannelInterface $logger_channel, ImmutableConfig $config) {
    $this->client = $client;
    $this->loggerChannel = $logger_channel;
    $this->config = $config;
  }

  /**
   * {@inheritdoc}
   */
  public function indexDataSet(DatasetInterface $dataset): bool {
    try {
      $index_params = [];
      $index_params_base = [];
      $index_name = $this->config->get('index_name_prefix') . $dataset->uuid();
      $this->createIndex($index_name);
      $index_params_base['index'] = $index_name;
      foreach ($dataset->data() as $delta => $data) {
        $index_params['id'] = $dataset->uuid() . ':' . $delta;
        assert($data instanceof DatasetData);
        $index_params['body'] = $data->getArrayCopy();
        $this->client->index($index_params_base + $index_params);
      }
      $dataset->setIndexingComplete();
      return TRUE;
    }
    catch (\Exception $e) {
      $this->loggerChannel->critical("The dataset data could not be indexed due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    return FALSE;
  }

  /**
   * A method to create the index (if it doesn't exist).
   *
   * @param string $name
   *  The name of the index.
   *
   * @throws \Drupal\data_pipelines\Index\CouldNotCreateIndexException
   *   When the index can't be created.
   */
  protected function createIndex(string $name): void {
    try {
      $index_params = [];
      $index_params_base = [];
      $index_params_base['index'] = $name;
      $index_params['body'] = [
        'settings' => [
          'number_of_shards' => self::INDEX_SHARDS,
          'number_of_replicas' => self::INDEX_REPLICAS,
          'codec' => self::INDEX_CODEC,
        ],
      ];
      if ($this->client->indices()->exists($index_params_base)) {
        return;
      }
      $response = $this->client->indices()->create($index_params_base + $index_params);
      if ($this->client->CheckResponseAck($response)) {
        return;
      }
    }
    catch (\Exception $e) {
      $this->loggerChannel->critical("The dataset index could not be created due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    // If we got this far, the index could not be created.
    throw new CouldNotCreateIndexException();
  }

  /**
   * {@inheritdoc}
   */
  public function deleteDataSetFromIndex(IndexingRequest $request): void {
    try {
      $this->client->indices()->delete([
        'index' => $request->getUuid(),
      ]);
    }
    catch (\Exception $e) {
      $this->loggerChannel->critical("The dataset @label could not be removed from the index due to @message", [
        '@message' => $e->getMessage(),
        '@label' => $request->getLabel(),
      ]);
      throw DatasetNotRemovedFromIndexException::fromException($e, $request->getLabel());
    }
  }

}
